(* Run provers in parallel on multiple problems. *)

let load_atpdeps fname =
  let go probs line = match Utils.string_split_on ':' line with
      [t; e] -> (t, Utils.string_split_on ' ' e) :: probs
    | [t] -> (t, []) :: probs
    | _ -> failwith "read atp" in

  Utils.foldl_file go [] fname;;


(* Average number of dependencies per problem. *)
let avg_n_deps probs =
  let depslen (_, deps) = List.length deps in
  let total = List.fold_left (fun acc p -> acc + depslen p) 0 probs in
  float_of_int total /. float_of_int (List.length probs)
;;


let print_problem (prob, deps) =
  print_endline "To show:"; print_thm prob; print_newline ();
  print_endline "Premises:";
  List.iter (fun t -> print_thm t; print_newline ()) deps;
  print_newline ();;


(* try to prove a problem with a given prover *)
let prove_prob prover (probh, depsh) =
  let findn = find_thm o find_hash in
  let (prob, deps) = (findn probh, List.map findn depsh) in
  (*print_problem (prob, deps);*)
  try Some (run_prover prover deps (concl prob))
  with
      Failure f -> print_string f; None
    | _ -> None
;;

(* write theorem with a given problem hash to a new file in dir *)
let write_thm dir (probh, _) th =
  let oc = open_out (dir ^ "/" ^ probh ^ ".p") in
  Printf.fprintf oc "%s%!" (lstring_of_thm th);
  close_out oc
;;

(* if there are problems left, start new thread and return remaining probs *)
let maybe_start prover dir = function
  ([], _) -> None
| (prob :: probs, n) ->
    Printf.printf "Start problem %i (%s)\n" n (string_of_prover prover);
    (match Unix.fork () with
      0 ->
       (match prove_prob prover prob with
          Some th -> write_thm dir prob th
        | None -> ());
        exit 0
    | pid -> Some (pid, (probs, n+1)))
;;

(* run a single prover on a set of problems, writing proven problems to dir *)
let process (Utils.Timeout t as timeout) n_processes probs base_dir prover =
  let p = string_of_prover prover in
  let dir = base_dir ^ "/" ^ p in
  Unix.mkdir dir 0o755;

  let start_t = Unix.gettimeofday ()
  and (_,n) = Utils.parallelise (maybe_start prover dir)
    (probs, 0) n_processes timeout
  and end_t = Unix.gettimeofday () in

  let oc = open_out (dir ^ ".txt") in
  Printf.fprintf oc "Statistics for prover %s, timeout %f seconds:\n" p t;
  Printf.fprintf oc "Time consumed: %f seconds\n" (end_t -. start_t);
  Printf.fprintf oc "Number of problems: %i\n" n;
  Printf.fprintf oc "Avg. number of deps: %f\n" (avg_n_deps probs);
  Printf.fprintf oc "Problems solved: %i\n" (Array.length (Sys.readdir dir));
  close_out oc
;;

let run_provers timeout n_processes probs dir =
  List.iter (process timeout n_processes probs dir);;

