module Leancop = struct

(* ----------------------------------------------------------------------- *)
(* Leancop parameters.                                                     *)
(* ----------------------------------------------------------------------- *)

(* If cut is disabled, Leancop will be complete, but slow.
   Enabling cut makes Leancop incomplete, but usually also makes it
   prove more problems in the same amount of time.
 *)
let cut = ref true;;



let hash_def = new_definition `hashek = T`;;
let hash_prop = prove(`((x /\ hashek) ==> (y /\ hashek)) ==> (x ==> y)`, SIMP_TAC[hash_def]);;
let hash_eq = prove (`~ hashek <=> F`, ASM_SIMP_TAC[hash_def; NOT_CLAUSES]);;


let db = ref ([] : (int * ((Meson.fol_atom list * (int * Meson.fol_term list)) * (int * thm) * bool) list) list);;

let rec inst_tm_stop cont env tm = match tm with
    Meson.Fvar(v) -> if mem v cont then (try inst_tm_stop cont env (rev_assoc v env) with Failure _ -> tm) else tm
  | Meson.Fnapp(f,args) -> Meson.Fnapp (f,map (inst_tm_stop cont env) args);;

let inc_vars cont subst off (x, l) =
  let rec inc_vs = function
    Meson.Fvar x -> inst_tm_stop cont subst (Meson.Fvar (x + off))
  | Meson.Fnapp (x, l) -> Meson.Fnapp (x, map inc_vs l) in
  (x, map inc_vs l);;

(* Immediately substitute new introduced and substituted variables *)
let optimize_subst off (nsub, nv) rest =
  if nv = [] then (nsub, map (inc_vars [] [] off) rest) else
  let nrest = List.map (inc_vars nv nsub off) rest in
  let nnsub = filter (fun (_, k) -> k < off or not (mem k nv)) nsub in
  let nnsub = map (fun (t, k) -> ((if mem k nv then inst_tm_stop nv nsub t else t), k)) nnsub in
  nnsub, nrest
;;


(* ----------------------------------------------------------------------- *)
(* Modified functions from Meson.                                          *)
(* ----------------------------------------------------------------------- *)

let rec fol_unify offset tm1 tm2 ((env, nv) as sofar) =
  match tm1,tm2 with
    Meson.Fnapp(f,fargs),Meson.Fnapp(g,gargs) ->
      if f <> g then failwith "" else
        itlist2 (fol_unify offset) fargs gargs sofar
  | _,Meson.Fvar(x) ->
    (let x' = x + offset in
     try let tm2' = rev_assoc x' env in
         fol_unify 0 tm1 tm2' sofar
     with Failure "find" ->
       if Meson.istriv env x' tm1 then sofar
       else ((tm1,x')::env, x' :: nv))
  | Meson.Fvar(x),_ ->
    (try let tm1' = rev_assoc x env in
         fol_unify offset tm1' tm2 sofar
     with Failure "find" ->
       let tm2' = Meson.fol_subst_bump offset [] tm2 in
       if Meson.istriv env x tm2' then sofar
       else ((tm2',x)::env, x :: nv))

let fol_unify_l off l1 l2 sf =
  List.fold_left (fun sf (t1, t2) -> fol_unify off t1 t2 sf) sf (zip l1 l2)

let fol_unify_lit offset (h1, l1) (h2, l2) env =
  if h1 <> h2 then failwith "" else
  List.fold_left (fun sf (t1, t2) -> fol_unify offset t1 t2 sf) (env, []) (zip l1 l2)

let PREPARE_LEAN_TAC ths =
  (* mostly from Meson *)
  REFUTE_THEN (LABEL_TAC "leanconcl") THEN
  Meson.POLY_ASSUME_TAC (map GEN_ALL ths) THEN

  (* Leancop-specific part *)
  USE_THEN "leanconcl" (fun th -> UNDISCH_TAC (concl th)) THEN
  REWRITE_TAC [TAUT `!t. (~t ==> F <=> t)`] THEN
  POP_ASSUM_LIST (fun l -> ASSUME_TAC (if l = [] then TRUTH else end_itlist CONJ l)) THEN
  W(MAP_EVERY(UNDISCH_TAC o concl o snd) o fst) THEN
  MATCH_MP_TAC hash_prop THEN
  DISCH_THEN (fun th -> EVERY (map ASSUME_TAC (CONJUNCTS th))) THEN
  REFUTE_THEN ASSUME_TAC THEN

    (* unmodified Meson parts start here *)
    W(MAP_EVERY(UNDISCH_TAC o concl o snd) o fst) THEN
    SELECT_ELIM_TAC THEN
    W(fun (asl,w) -> MAP_EVERY (fun v -> SPEC_TAC(v,v)) (frees w)) THEN
    CONV_TAC(PRESIMP_CONV THENC
             TOP_DEPTH_CONV BETA_CONV THENC
             LAMBDA_ELIM_CONV THENC
             CONDS_CELIM_CONV THENC
             Meson.QUANT_BOOL_CONV) THEN
    REPEAT(GEN_TAC ORELSE DISCH_TAC) THEN
    REFUTE_THEN ASSUME_TAC THEN
    RULE_ASSUM_TAC(CONV_RULE(NNF_CONV THENC SKOLEM_CONV)) THEN
    REPEAT (FIRST_X_ASSUM CHOOSE_TAC) THEN
    ASM_FOL_TAC THEN
(*  SPLIT_TAC (!meson_split_limit) THEN *)
    RULE_ASSUM_TAC(CONV_RULE(PRENEX_CONV THENC WEAK_CNF_CONV)) THEN
    RULE_ASSUM_TAC(repeat
     (fun th -> SPEC(genvar(type_of(fst(dest_forall(concl th))))) th)) THEN
    REPEAT (FIRST_X_ASSUM (Meson.CONJUNCTS_THEN' ASSUME_TAC)) THEN
    RULE_ASSUM_TAC(CONV_RULE(ASSOC_CONV DISJ_ASSOC)) THEN
    REPEAT (FIRST_X_ASSUM SUBST_VAR_TAC)


(* ----------------------------------------------------------------------- *)
(* Leancop.                                                                *)
(* ----------------------------------------------------------------------- *)

let LIT off sub (p, tl) lim fn =
  if lim < 0 then () else
  let db = try assoc p !db with Failure _ -> [] in
  let iter_fun ((rest, (_, dbtl)), th, ground) =
    try
      if ground then
        let s, _ = fol_unify_l 0 tl dbtl (sub, []) in fn off s rest lim th (p, tl)
      else
        let noff = Meson.offinc + off in
        let (ns, nv) = fol_unify_l noff tl dbtl (sub, []) in
        let s, rest = optimize_subst noff (ns, nv) rest in fn noff s rest (lim - 1) th (p, tl)
    with Failure _ -> () in
  List.iter iter_fun db;;

type proof =
  Lem of Meson.fol_atom
| Pat of Meson.fol_atom
| Res of Meson.fol_atom * (int * thm);;

exception Solved of (Meson.fol_term * int) list * proof list;;
exception Lcut of int;;

let rec lprove off subst path limit lemmas todo = function
  [] -> begin match todo with
      [] -> raise (Solved (subst, []))
    | (ct,p,pl,l,c) :: t -> lprove off subst p pl l t c; if ct > 0 then raise (Lcut ct)
    end
| ((lit1 :: rest_clause) as clause) ->

      if (exists (fun x -> exists (Meson.fol_atom_eq subst x) path)) clause then () else

      if exists (Meson.fol_atom_eq subst lit1) lemmas && ((try lprove off subst path limit lemmas todo rest_clause with Solved (s,x) -> raise (Solved (s,(Lem (lit1)) :: x))); !cut) then () else

      let neglit = Meson.mk_negated lit1 in
      if itlist (fun plit sf -> if sf then true else
        try let (s, _) = fol_unify_lit 0 neglit plit subst in
        ((try lprove off s path limit (lit1 :: lemmas) todo rest_clause with Solved (s,x) -> raise (Solved (s,(Pat (lit1)) :: x))); true) with Failure _ -> sf) path false && !cut then () else

      (* lit1 rather than neglit in MESON infra *)
      try LIT off subst lit1 limit (fun noff s cla1 new_limit th lit1 ->
        let todo = ((if !cut then limit else -1), path, limit, lit1 :: lemmas, rest_clause) :: todo in
        try lprove noff s (lit1 :: path) new_limit lemmas todo cla1 with Solved (s,x) -> raise (Solved (s,Res (lit1, th) :: x))
      ) with Lcut n -> if limit = n then () else raise (Lcut n);;

let hol_negate tm =
      try dest_neg tm with Failure _ -> mk_neg tm;;

let ss s (x, l) = (x, List.map (Meson.fol_subst_partial s) l);;
let ssl s = List.map (fun (x, l) -> (x, List.map (Meson.fol_subst_partial s) l));;

let rec recon sub lemmas = function
    Pat g :: tl -> (ASSUME (Meson.hol_of_literal (ss sub g)), tl)
  | Lem g :: tl -> find (fun x -> concl x = Meson.hol_of_literal (ss sub g)) lemmas, tl
  | Res (g, (n, th)) :: tl ->
    let foldf (found, lms, prf) _ =
      let (nthm, nprf) = recon sub lms prf in (nthm :: found, nthm :: lms, nprf) in
    let (ths, nlms, nprf) = List.fold_left foldf ([],lemmas,tl) (List.tl (disjuncts (concl th))) in
    let cth = Meson.make_hol_contrapos(n,th) in
    let hth = if ths = [] then cth else MATCH_MP cth (end_itlist CONJ (rev ths)) in
    let ith = PART_MATCH I hth (Meson.hol_of_literal (ss sub g)) in
    Meson.finish_RULE (DISCH (hol_negate(concl ith)) ith), nprf
  | [] -> assert false
;;

let rec fol_ground = function
    Meson.Fvar _ -> false
  | Meson.Fnapp (_, l) -> forall fol_ground l;;
let at_ground (_, l) = forall fol_ground l;;

let add_ground =
  List.map (fun (i, info) -> (i, List.map
    (fun  ((al, a), t) -> ((al, a), t, false && at_ground a && forall at_ground al)) info
  ));;

let rec LEANCOP_DEEPEN ths =
  Meson.clear_contrapos_cache(); Meson.reset_vars(); Meson.reset_consts();
  let ths' = ths @ Meson.create_equality_axioms (map concl ths) in
  db := add_ground (map (fun (a, l) -> a, rev l) (Meson.fol_of_hol_clauses ths'));
  let hashf = Meson.fol_of_literal [] [] `hashek` in
  let hashfn = Meson.mk_negated hashf in
  let rec deepen i = lprove Meson.offinc [] [] i [] [] [hashfn]; deepen (i+1) in
  let (s, l) = try deepen 1 with Solved (s,l) -> (s,l) in
  EQ_MP hash_eq (fst (recon s [] l))
;;

let PURE_LEANCOP gl =
  (FIRST_ASSUM CONTR_TAC ORELSE W(ACCEPT_TAC o LEANCOP_DEEPEN o map snd o fst)) gl;;

let LEANCOP_TAC ths = PREPARE_LEAN_TAC ths THEN PURE_LEANCOP;;
let LEANCOP ths tm = prove(tm, LEANCOP_TAC ths);;

end
