let map = List.map;;

let var_name = string_of_int;;

let rec transTerm = function
  Tptp.V var -> Term.Var (var_name var)
| Tptp.A (name, terms) -> Term.Fn (name, map transTerm terms);;

let rec transFormula = function
  Tptp.Atom (name, terms) -> Formula.Atom (name, map transTerm terms)
| Tptp.Neg form -> Formula.Not (transFormula form)
| Tptp.Conj (f1, f2) -> Formula.And (transFormula f1, transFormula f2)
| Tptp.Disj (f1, f2) -> Formula.Or  (transFormula f1, transFormula f2)
| Tptp.Eqiv (f1, f2) -> Formula.Iff (transFormula f1, transFormula f2)
| Tptp.Forall (var, form) -> Formula.Forall (var_name var, transFormula form)
| Tptp.Exists (var, form) -> Formula.Exists (var_name var, transFormula form);;

let isConjecture = function
  (_, _, "conjecture", _) -> true
| _ -> false;;


let tptp_to_metis forms =
  let (conjecture, axioms) = List.partition isConjecture forms in

  let axiomatize (_, _, _, form) = Normalize.mkAxiom (transFormula form) in
  let cnfify ths = map Thm.axiom (map fst (Normalize.proveCnf ths)) in
  let conjecture = cnfify (map axiomatize conjecture)
  and axioms     = cnfify (map axiomatize axioms) in

  (axioms, conjecture)
;;

let rec resolution_loop limit res = let open Resolution in
  match iterate res with
    Decided dec -> Some dec
  | Undecided res -> if limit () then None else resolution_loop limit res
;;

let new_limit t =
  let start = Sys.time () in
  let check () =
    let now = Sys.time () in
    now >= start +. t
  in check
;;

let work fname =
  let f = Tptp_read.tptp_file fname in
  let (ax, cj) = tptp_to_metis f in let open Ax_cj in let open Resolution in
  let res = newResolution default {axioms_thm = ax; conjecture_thm = cj} in
  let limit = new_limit 2.0 in
  match resolution_loop limit res with
    None -> print_endline "timeout"
  | Some result -> match result with
        Contradiction _ -> print_endline "contradiction"
      | Satisfiable sat ->
        print_endline "satisfiable:";
        List.iter (fun t -> print_endline (Thm.toString t)) sat
;;

let main () =
  print_endline "Welcome to OCaml Metis.";
  for i = 1 to Array.length Sys.argv - 1 do
    Printf.printf "Working on: %s ... " Sys.argv.(i);
    try
      work Sys.argv.(i)
    with Failure f -> print_endline f
  done
;;

let _ = main ();;
