(* ========================================================================= *)
(* NORMALIZING FORMULAS                                                      *)
(* Copyright (c) 2001 Joe Leslie-Hurd, distributed under the MIT license     *)
(* ========================================================================= *)

open Useful;;
open Order
open Real

(* ------------------------------------------------------------------------- *)
(* Constants.                                                                *)
(* ------------------------------------------------------------------------- *)

let prefix = "FOFtoCNF";;

let skolemPrefix = "skolem" ^ prefix;;

let definitionPrefix = "definition" ^ prefix;;

(* ------------------------------------------------------------------------- *)
(* Storing huge real numbers as their log.                                   *)
(* ------------------------------------------------------------------------- *)

type logReal = Log_real of real;;

let compareLogReal (Log_real logX, Log_real logY) =
    Real.compare (logX,logY);;

let zeroLogReal = Log_real (-1.0);;

let oneLogReal = Log_real 0.0;;

  let isZero logX = logX < 0.0;;

  (* Assume logX >= logY >= 0.0 *)
  let add logX logY = logX +. Math.ln (1.0 +. Math.exp (logY -. logX));;

  let isZeroLogReal (Log_real logX) = isZero logX;;

  let multiplyLogReal (Log_real logX) (Log_real logY) =
      if isZero logX or isZero logY then zeroLogReal
      else Log_real (logX +. logY);;

  let addLogReal (Log_real logX as lx) (Log_real logY as ly) =
      if isZero logX then ly
      else if isZero logY then lx
      else if logX < logY then Log_real (add logY logX)
      else Log_real (add logX logY);;

  let withinRelativeLogReal logDelta (Log_real logX) (Log_real logY) =
      isZero logX or
      (not (isZero logY) && logX < logY +. logDelta);;


(* ------------------------------------------------------------------------- *)
(* Counting the clauses that would be generated by conjunctive normal form.  *)
(* ------------------------------------------------------------------------- *)

let countLogDelta = 0.01;;

type count = {positive : logReal; negative : logReal};;

let countCompare (count1,count2) =
      let {positive = p1} = count1
      and {positive = p2} = count2
    in
      compareLogReal (p1,p2)
    ;;

let countNegate ({positive = p; negative = n}) =
    {positive = n; negative = p};;

let countLeqish count1 count2 =
      let {positive = p1} = count1
      and {positive = p2} = count2
    in
      withinRelativeLogReal countLogDelta p1 p2
    ;;

(*MetisDebug
let countEqualish count1 count2 =
    countLeqish count1 count2 &&
    countLeqish count2 count1;;

let countEquivish count1 count2 =
    countEqualish count1 count2 &&
    countEqualish (countNegate count1) (countNegate count2);;
*)

let countTrue = {positive = zeroLogReal; negative = oneLogReal};;

let countFalse = {positive = oneLogReal; negative = zeroLogReal};;

let countLiteral = {positive = oneLogReal; negative = oneLogReal};;

let countAnd2 (count1,count2) =
      let {positive = p1; negative = n1} = count1
      and {positive = p2; negative = n2} = count2
      in let p = addLogReal p1 p2
      and n = multiplyLogReal n1 n2
    in
      {positive = p; negative = n}
    ;;

let countOr2 (count1,count2) =
      let {positive = p1; negative = n1} = count1
      and {positive = p2; negative = n2} = count2
      in let p = multiplyLogReal p1 p2
      and n = addLogReal n1 n2
    in
      {positive = p; negative = n}
    ;;

(* Whether countXor2 is associative or not is an open question. *)

let countXor2 (count1,count2) =
      let {positive = p1; negative = n1} = count1
      and {positive = p2; negative = n2} = count2
      in let p = addLogReal (multiplyLogReal p1 p2) (multiplyLogReal n1 n2)
      and n = addLogReal (multiplyLogReal p1 n2) (multiplyLogReal n1 p2)
    in
      {positive = p; negative = n}
    ;;

let countDefinition body_count = countXor2 (countLiteral,body_count);;


(* ------------------------------------------------------------------------- *)
(* A type of normalized formula.                                             *)
(* ------------------------------------------------------------------------- *)

type formula =
    True
  | False
  | Literal of Name.Set.set * Literal.literal
  | And of Name.Set.set * count * formula Pset.set
  | Or of Name.Set.set * count * formula Pset.set
  | Xor of Name.Set.set * count * bool * formula Pset.set
  | Exists of Name.Set.set * count * Name.Set.set * formula
  | Forall of Name.Set.set * count * Name.Set.set * formula;;

let mkExists (x, y, z, a) = Exists (x,y,z,a);;
let mkForall (x, y, z, a) = Forall (x,y,z,a);;

let rec compare f1_f2 =
    if Portable.pointerEqual f1_f2 then Equal
    else
      match f1_f2 with
        (True,True) -> Equal
      | (True,_) -> Less
      | (_,True) -> Greater
      | (False,False) -> Equal
      | (False,_) -> Less
      | (_,False) -> Greater
      | (Literal (_,l1), Literal (_,l2)) -> Literal.compare (l1,l2)
      | (Literal _, _) -> Less
      | (_, Literal _) -> Greater
      | (And (_,_,s1), And (_,_,s2)) -> Pset.compare (s1,s2)
      | (And _, _) -> Less
      | (_, And _) -> Greater
      | (Or (_,_,s1), Or (_,_,s2)) -> Pset.compare (s1,s2)
      | (Or _, _) -> Less
      | (_, Or _) -> Greater
      | (Xor (_,_,p1,s1), Xor (_,_,p2,s2)) ->
        (match boolCompare (p1,p2) with
           Less -> Less
         | Equal -> Pset.compare (s1,s2)
         | Greater -> Greater)
      | (Xor _, _) -> Less
      | (_, Xor _) -> Greater
      | (Exists (_,_,n1,f1), Exists (_,_,n2,f2)) ->
        (match Name.Set.compare (n1,n2) with
           Less -> Less
         | Equal -> compare (f1,f2)
         | Greater -> Greater)
      | (Exists _, _) -> Less
      | (_, Exists _) -> Greater
      | (Forall (_,_,n1,f1), Forall (_,_,n2,f2)) ->
        (match Name.Set.compare (n1,n2) with
           Less -> Less
         | Equal -> compare (f1,f2)
         | Greater -> Greater);;

let empty = Pset.empty compare;;

let singleton = Pset.singleton compare;;

  let rec neg = function
      True -> False
    | False -> True
    | (Literal (fv,lit)) -> Literal (fv, Literal.negate lit)
    | (And (fv,c,s)) -> Or (fv, countNegate c, neg_set s)
    | (Or (fv,c,s)) -> And (fv, countNegate c, neg_set s)
    | (Xor (fv,c,p,s)) -> Xor (fv, c, not p, s)
    | (Exists (fv,c,n,f)) -> Forall (fv, countNegate c, n, neg f)
    | (Forall (fv,c,n,f)) -> Exists (fv, countNegate c, n, neg f)

  and neg_set s = Pset.foldl neg_elt empty s

  and neg_elt (f,s) = Pset.add s (neg f);;

  let negate = neg;;

  let negateSet = neg_set;;

let negateMember x s = Pset.member (negate x) s;;

let negateDisjoint s1 s2 =
  let member s x = negateMember x s in
      if Pset.size s1 < Pset.size s2 then not (Pset.exists (member s2) s1)
      else not (Pset.exists (member s1) s2);;

let polarity = function
    True -> true
  | False -> false
  | (Literal (_,(pol,_))) -> not pol
  | (And _) -> true
  | (Or _) -> false
  | (Xor (_,_,pol,_)) -> pol
  | (Exists _) -> true
  | (Forall _) -> false;;

(*MetisDebug
let polarity = fun f ->
    let
      let res1 = compare (f, negate f) = Less
      let res2 = polarity f
      let _ = res1 = res2 or raise Bug "polarity"
    in
      res2
    end;;
*)

let applyPolarity b fm = match b with
    true -> fm
  | false -> negate fm;;

let freeVars = function
    True -> Name.Set.empty
  | False -> Name.Set.empty
  | (Literal (fv,_)) -> fv
  | (And (fv,_,_)) -> fv
  | (Or (fv,_,_)) -> fv
  | (Xor (fv,_,_,_)) -> fv
  | (Exists (fv,_,_,_)) -> fv
  | (Forall (fv,_,_,_)) -> fv;;

let freeIn v fm = Name.Set.member v (freeVars fm);;

let freeVarsSet =
      let free (fm,acc) = Name.Set.union (freeVars fm) acc
    in
      Pset.foldl free Name.Set.empty
    ;;

let count = function
    True -> countTrue
  | False -> countFalse
  | (Literal _) -> countLiteral
  | (And (_,c,_)) -> c
  | (Or (_,c,_)) -> c
  | (Xor (_,c,p,_)) -> if p then c else countNegate c
  | (Exists (_,c,_,_)) -> c
  | (Forall (_,c,_,_)) -> c;;

let countAndSet =
      let countAnd (fm,c) = countAnd2 (count fm, c)
    in
      Pset.foldl countAnd countTrue
    ;;

let countOrSet =
      let countOr (fm,c) = countOr2 (count fm, c)
    in
      Pset.foldl countOr countFalse
    ;;

let countXorSet =
      let countXor (fm,c) = countXor2 (count fm, c)
    in
      Pset.foldl countXor countFalse
    ;;

let and2 = function
    (False,_) -> False
  | (_,False) -> False
  | (True,f2) -> f2
  | (f1,True) -> f1
  | (f1,f2) ->
      let (fv1,c1,s1) =
          (match f1 with
            And (fv,c,s) -> (fv,c,s)
          | _ -> (freeVars f1, count f1, singleton f1))

      and (fv2,c2,s2) =
          (match f2 with
            And (fv,c,s) -> (fv,c,s)
          | _ -> (freeVars f2, count f2, singleton f2))
    in
      if not (negateDisjoint s1 s2) then False
      else
          let s = Pset.union s1 s2
        in
          (match Pset.size s with
            0 -> True
          | 1 -> Pset.pick s
          | n ->
            if n = Pset.size s1 + Pset.size s2 then
              And (Name.Set.union fv1 fv2, countAnd2 (c1,c2), s)
            else
              And (freeVarsSet s, countAndSet s, s))
    ;;

let fAndList = Mlist.foldl and2 True;;

let fAndSet = Pset.foldl and2 True;;

let or2 = function
    (True,_) -> True
  | (_,True) -> True
  | (False,f2) -> f2
  | (f1,False) -> f1
  | (f1,f2) ->
      let (fv1,c1,s1) =
          (match f1 with
            Or (fv,c,s) -> (fv,c,s)
          | _ -> (freeVars f1, count f1, singleton f1))

      and (fv2,c2,s2) =
          (match f2 with
            Or (fv,c,s) -> (fv,c,s)
          | _ -> (freeVars f2, count f2, singleton f2))
    in
      if not (negateDisjoint s1 s2) then True
      else
          let s = Pset.union s1 s2
        in
          (match Pset.size s with
            0 -> False
          | 1 -> Pset.pick s
          | n ->
            if n = Pset.size s1 + Pset.size s2 then
              Or (Name.Set.union fv1 fv2, countOr2 (c1,c2), s)
            else
              Or (freeVarsSet s, countOrSet s, s))
    ;;

let fOrList = Mlist.foldl or2 False;;

let fOrSet = Pset.foldl or2 False;;

let pushOr2 (f1,f2) =
      let s1 = match f1 with And (_,_,s) -> s | _ -> singleton f1
      and s2 = match f2 with And (_,_,s) -> s | _ -> singleton f2

      in let g x1 (x2,acc) = and2 (or2 (x1,x2), acc)

      in let f (x1,acc) = Pset.foldl (g x1) acc s2
    in
      Pset.foldl f True s1
    ;;

let pushOrList = Mlist.foldl pushOr2 False;;

  let normalize fm =
        let p = polarity fm
        in let fm = applyPolarity p fm
      in
        (freeVars fm, count fm, p, singleton fm)
      ;;

  let fXor2 = function
      (False,f2) -> f2
    | (f1,False) -> f1
    | (True,f2) -> negate f2
    | (f1,True) -> negate f1
    | (f1,f2) ->
        let (fv1,c1,p1,s1) = match f1 with Xor (x1,x2,x3,x4) -> (x1,x2,x3,x4) | _ -> normalize f1
        and (fv2,c2,p2,s2) = match f2 with Xor (x1,x2,x3,x4) -> (x1,x2,x3,x4) | _ -> normalize f2

        in let s = Pset.symmetricDifference s1 s2

        in let fm =
            match Pset.size s with
              0 -> False
            | 1 -> Pset.pick s
            | n ->
              if n = Pset.size s1 + Pset.size s2 then
                Xor (Name.Set.union fv1 fv2, countXor2 (c1,c2), true, s)
              else
                Xor (freeVarsSet s, countXorSet s, true, s)

        in let p = p1 = p2
      in
        applyPolarity p fm
      ;;

let fXorList = Mlist.foldl fXor2 False;;

let fXorSet = Pset.foldl fXor2 False;;

let fXorPolarityList (p,l) = applyPolarity p (fXorList l);;

let fXorPolaritySet (p,s) = applyPolarity p (fXorSet s);;

let destXor = function
    (Xor (_,_,p,s)) ->
      let (fm1,s) = Pset.deletePick s
      in let fm2 =
          if Pset.size s = 1 then applyPolarity p (Pset.pick s)
          else Xor (freeVarsSet s, countXorSet s, p, s)
    in
      (fm1,fm2)
  | _ -> raise (Error "destXor");;

let pushXor fm =
      let (f1,f2) = destXor fm
      in let f1' = negate f1
      and f2' = negate f2
    in
      and2 (or2 (f1,f2), or2 (f1',f2'))
    ;;

let exists1 (v,init_fm) =
      let exists_gen fm =
            let fv = Name.Set.delete (freeVars fm) v
            in let c = count fm
            in let n = Name.Set.singleton v
          in
            Exists (fv,c,n,fm)

      in let rec exists fm = if freeIn v fm then exists_free fm else fm

      and exists_free = function
          (Or (_,_,s)) -> fOrList (Pset.transform exists s)
        | (And (_,_,s) as fm) ->
            let sv = Pset.filter (freeIn v) s
          in
            if Pset.size sv <> 1 then exists_gen fm
            else
                let fm = Pset.pick sv
                in let s = Pset.delete s fm
              in
                and2 (exists_free fm, fAndSet s)
        | (Exists (fv,c,n,f)) ->
          Exists (Name.Set.delete fv v, c, Name.Set.add n v, f)
        | fm -> exists_gen fm
    in
      exists init_fm
    ;;

let fExistsList (vs,f) = Mlist.foldl exists1 f vs;;

let fExistsSet (n,f) = Name.Set.foldl exists1 f n;;

let forall1 (v,init_fm) =
      let forall_gen fm =
            let fv = Name.Set.delete (freeVars fm) v
            in let c = count fm
            in let n = Name.Set.singleton v
          in
            Forall (fv,c,n,fm)

      in let rec forall fm = if freeIn v fm then forall_free fm else fm

      and forall_free = function
          (And (_,_,s)) -> fAndList (Pset.transform forall s)
        | (Or (_,_,s) as fm) ->
            let sv = Pset.filter (freeIn v) s
          in
            if Pset.size sv <> 1 then forall_gen fm
            else
                let fm = Pset.pick sv
                in let s = Pset.delete s fm
              in
                or2 (forall_free fm, fOrSet s)
        | (Forall (fv,c,n,f)) ->
          Forall (Name.Set.delete fv v, c, Name.Set.add n v, f)
        | fm -> forall_gen fm
    in
      forall init_fm
    ;;

let fForallList (vs,f) = Mlist.foldl forall1 f vs;;

let fForallSet (n,f) = Name.Set.foldl forall1 f n;;

let generalize f = fForallSet (freeVars f, f);;

  let subst_fv fvSub =
        let add_fv (v,s) = Name.Set.union (Name.Map.get fvSub v) s
      in
        Name.Set.foldl add_fv Name.Set.empty
      ;;

  let subst_rename (v,(avoid,bv,sub,domain,fvSub)) =
        let v' = Term.variantPrime avoid v
        in let avoid = Name.Set.add avoid v'
        in let bv = Name.Set.add bv v'
        in let sub = Substitute.insert sub (v, Term.Var v')
        in let domain = Name.Set.add domain v
        in let fvSub = Name.Map.insert fvSub (v, Name.Set.singleton v')
      in
        (avoid,bv,sub,domain,fvSub)
      ;;

  let rec subst_check sub domain fvSub fm =
        let domain = Name.Set.intersect domain (freeVars fm)
      in
        if Name.Set.null domain then fm
        else subst_domain sub domain fvSub fm

  and subst_domain sub domain fvSub fm =
      match fm with
        Literal (fv,lit) ->
          let fv = Name.Set.difference fv domain
          in let fv = Name.Set.union fv (subst_fv fvSub domain)
          in let lit = Literal.subst sub lit
        in
          Literal (fv,lit)
      | And (_,_,s) ->
        fAndList (Pset.transform (subst_check sub domain fvSub) s)
      | Or (_,_,s) ->
        fOrList (Pset.transform (subst_check sub domain fvSub) s)
      | Xor (_,_,p,s) ->
        fXorPolarityList (p, Pset.transform (subst_check sub domain fvSub) s)
      | Exists (fv,c,n,f) -> subst_quant mkExists sub domain fvSub (fv,c,n,f)
      | Forall (fv,c,n,f) -> subst_quant mkForall sub domain fvSub (fv,c,n,f)
      | _ -> raise (Bug "subst_domain")

  and subst_quant quant sub domain fvSub (fv,c,bv,fm) =
        let sub_fv = subst_fv fvSub domain
        in let fv = Name.Set.union sub_fv (Name.Set.difference fv domain)
        in let captured = Name.Set.intersect bv sub_fv
        in let bv = Name.Set.difference bv captured
        in let avoid = Name.Set.union fv bv
        in let (_,bv,sub,domain,fvSub) =
            Name.Set.foldl subst_rename (avoid,bv,sub,domain,fvSub) captured
        in let fm = subst_domain sub domain fvSub fm
      in
        quant (fv,c,bv,fm)
      ;;

  let subst sub =
        let mk_dom (v,tm,(d,fv)) =
            (Name.Set.add d v, Name.Map.insert fv (v, Term.freeVars tm))

        in let domain_fvSub = (Name.Set.empty, Name.Map.newMap ())
        in let (domain,fvSub) = Substitute.foldl mk_dom domain_fvSub sub
      in
        subst_check sub domain fvSub
      ;;

let rec fromFormula fm =
    match fm with
      Formula.True -> True
    | Formula.False -> False
    | Formula.Atom atm -> Literal (Atom.freeVars atm, (true,atm))
    | Formula.Not p -> negateFromFormula p
    | Formula.And (p,q) -> and2 (fromFormula p, fromFormula q)
    | Formula.Or (p,q) -> or2 (fromFormula p, fromFormula q)
    | Formula.Imp (p,q) -> or2 (negateFromFormula p, fromFormula q)
    | Formula.Iff (p,q) -> fXor2 (negateFromFormula p, fromFormula q)
    | Formula.Forall (v,p) -> forall1 (v, fromFormula p)
    | Formula.Exists (v,p) -> exists1 (v, fromFormula p)

and negateFromFormula fm =
    match fm with
      Formula.True -> False
    | Formula.False -> True
    | Formula.Atom atm -> Literal (Atom.freeVars atm, (false,atm))
    | Formula.Not p -> fromFormula p
    | Formula.And (p,q) -> or2 (negateFromFormula p, negateFromFormula q)
    | Formula.Or (p,q) -> and2 (negateFromFormula p, negateFromFormula q)
    | Formula.Imp (p,q) -> and2 (fromFormula p, negateFromFormula q)
    | Formula.Iff (p,q) -> fXor2 (fromFormula p, fromFormula q)
    | Formula.Forall (v,p) -> exists1 (v, negateFromFormula p)
    | Formula.Exists (v,p) -> forall1 (v, negateFromFormula p);;

  let lastElt (s : formula Pset.set) =
      match Pset.findr (kComb true) s with
        None -> raise (Bug "lastElt: empty set")
      | Some fm -> fm;;

  let negateLastElt s =
        let fm = lastElt s
      in
        Pset.add (Pset.delete s fm) (negate fm)
      ;;

  let rec form fm =
      match fm with
        True -> Formula.True
      | False -> Formula.False
      | Literal (_,lit) -> Literal.toFormula lit
      | And (_,_,s) -> Formula.listMkConj (Pset.transform form s)
      | Or (_,_,s) -> Formula.listMkDisj (Pset.transform form s)
      | Xor (_,_,p,s) ->
          let s = if p then negateLastElt s else s
        in
          Formula.listMkEquiv (Pset.transform form s)
      | Exists (_,_,n,f) -> Formula.listMkExists (Name.Set.toList n, form f)
      | Forall (_,_,n,f) -> Formula.listMkForall (Name.Set.toList n, form f);;

  let toFormula = form;;

let toLiteral = function
    (Literal (_,lit)) -> lit
  | _ -> raise (Error "Normalize.toLiteral");;

  let addLiteral (l,s) = Literal.Set.add s (toLiteral l);;
  let toClause = function
      False -> Literal.Set.empty
    | (Or (_,_,s)) -> Pset.foldl addLiteral Literal.Set.empty s
    | l -> Literal.Set.singleton (toLiteral l);;

(* ------------------------------------------------------------------------- *)
(* Negation normal form.                                                     *)
(* ------------------------------------------------------------------------- *)

let nnf fm = toFormula (fromFormula fm);;

(* ------------------------------------------------------------------------- *)
(* Basic conjunctive normal form.                                            *)
(* ------------------------------------------------------------------------- *)

  let skCounter : int Stringmap.map ref = ref (Stringmap.newMap ());;

  let newSkF n () =
        let m = !skCounter
        in let s = Name.toString n
        in let i = Option.getOpt (Stringmap.peek m s, 0)
        in let () = skCounter := Stringmap.insert m (s, i + 1)
        in let i = if i = 0 then "" else "_" ^ Int.toString i
        in let s = skolemPrefix ^ "_" ^ s ^ i
      in
        Name.fromString s
      ;;

  let newSkolemFunction n = Portable.critical (newSkF n) ();;

let skolemize fv bv fm =
      let fv = Name.Set.transform (fun x -> Term.Var x) fv

      in let mk (v,s) = Substitute.insert s (v, Term.Fn (newSkolemFunction v, fv))
    in
      subst (Name.Set.foldl mk Substitute.empty bv) fm
    ;;

  let rename avoid fv bv fm =
        let captured = Name.Set.intersect avoid bv
      in
        if Name.Set.null captured then fm
        else
            let ren (v,(a,s)) =
                  let v' = Term.variantPrime a v
                in
                  (Name.Set.add a v', Substitute.insert s (v, Term.Var v'))

            in let avoid = Name.Set.union (Name.Set.union avoid fv) bv

            in let (_,sub) = Name.Set.foldl ren (avoid,Substitute.empty) captured
          in
            subst sub fm
      ;;

  let rec cnfFm avoid fm =
(*MetisTrace5
      let
        let fm' = cnfFm' avoid fm
        let () = Print.trace pp "Normalize.cnfFm: fm" fm
        let () = Print.trace pp "Normalize.cnfFm: fm'" fm'
      in
        fm'
      end
  and cnfFm' avoid fm =
*)
      match fm with
        True -> True
      | False -> False
      | Literal _ -> fm
      | And (_,_,s) -> fAndList (Pset.transform (cnfFm avoid) s)
      | Or (fv,_,s) ->
          let avoid = Name.Set.union avoid fv
          in let (fms,_) = Pset.foldl cnfOr ([],avoid) s
        in
          pushOrList fms
      | Xor _ -> cnfFm avoid (pushXor fm)
      | Exists (fv,_,n,f) -> cnfFm avoid (skolemize fv n f)
      | Forall (fv,_,n,f) -> cnfFm avoid (rename avoid fv n f)

  and cnfOr (fm,(fms,avoid)) =
        let fm = cnfFm avoid fm
        in let fms = fm :: fms
        in let avoid = Name.Set.union avoid (freeVars fm)
      in
        (fms,avoid)
      ;;

  let basicCnf = cnfFm Name.Set.empty;;

(* ------------------------------------------------------------------------- *)
(* Finding the formula definition that minimizes the number of clauses.      *)
(* ------------------------------------------------------------------------- *)

  type best = count * formula option;;

  let rec minBreak countClauses fm best =
      match fm with
        True -> best
      | False -> best
      | Literal _ -> best
      | And (_,_,s) ->
        minBreakSet countClauses countAnd2 countTrue fAndSet s best
      | Or (_,_,s) ->
        minBreakSet countClauses countOr2 countFalse fOrSet s best
      | Xor (_,_,_,s) ->
        minBreakSet countClauses countXor2 countFalse fXorSet s best
      | Exists (_,_,_,f) -> minBreak countClauses f best
      | Forall (_,_,_,f) -> minBreak countClauses f best

  and minBreakSet countClauses count2 count0 mkSet fmSet best =
        let cumulatives fms =
              let fwd (fm,(c1,s1,l)) =
                    let c1' = count2 (count fm, c1)
                    and s1' = Pset.add s1 fm
                  in
                    (c1', s1', (c1,s1,fm) :: l)

              in let bwd ((c1,s1,fm),(c2,s2,l)) =
                    let c2' = count2 (count fm, c2)
                    and s2' = Pset.add s2 fm
                  in
                    (c2', s2', (c1,s1,fm,c2,s2) :: l)

              in let (c1,_,fms) = Mlist.foldl fwd (count0,empty,[]) fms
              in let (c2,_,fms) = Mlist.foldl bwd (count0,empty,[]) fms

(*MetisDebug
              let _ = countEquivish c1 c2 or
                      raise Bug ("cumulativeCounts: c1 = " ^ countToString c1 ^
                                 ", c2 = " ^ countToString c2)
*)
            in
              fms

        in let breakSing ((c1,_,fm,c2,_),best) =
              let cFms = count2 (c1,c2)

              in let countCls cFm = countClauses (count2 (cFms,cFm))
            in
              minBreak countCls fm best

        in let breakSet1 =
              let break c1 s1 fm c2 ((bcl,_) as best) =
                  if Pset.null s1 then best
                  else
                      let cDef = countDefinition (countXor2 (c1, count fm))
                      in let cFm = count2 (countLiteral,c2)
                      in let cl = countAnd2 (cDef, countClauses cFm)
                      in let noBetter = countLeqish bcl cl
                    in
                      if noBetter then best
                      else (cl, Some (mkSet (Pset.add s1 fm)))
            in
              fun ((c1,s1,fm,c2,s2),best) ->
                 break c1 s1 fm c2 (break c2 s2 fm c1 best)

        in let fms = Pset.toList fmSet

        in let breakSet measure best =
              let fms = sortMap (fun x -> measure (count x)) countCompare fms
            in
              Mlist.foldl breakSet1 best (cumulatives fms)

        in let best = Mlist.foldl breakSing best (cumulatives fms)
        in let best = breakSet iComb best
        in let best = breakSet countNegate best
        in let best = breakSet countClauses best
      in
        best;;

  let minimumDefinition fm =
        let cl = count fm
      in
        if countLeqish cl countLiteral then None
        else
            let (cl',def) = minBreak iComb fm (cl,None)
(*MetisTrace1
            let () =
                match def with
                  None -> ()
                | Some d ->
                  Print.trace pp ("defCNF: before = " ^ countToString cl ^
                                  ", after = " ^ countToString cl' ^
                                  ", definition") d
*)
          in
            def
      ;;

(* ------------------------------------------------------------------------- *)
(* Conjunctive normal form derivations.                                      *)
(* ------------------------------------------------------------------------- *)

type thm = Thm of formula * inference

and inference =
    Axiom of Formula.formula
  | Definition of string * Formula.formula
  | Simplify of thm * thm list
  | Conjunct of thm
  | Specialize of thm
  | Skolemize of thm
  | Clausify of thm;;

let parentsInference inf =
    match inf with
      Axiom _ -> []
    | Definition _ -> []
    | Simplify (th,ths) -> th :: ths
    | Conjunct th -> [th]
    | Specialize th -> [th]
    | Skolemize th -> [th]
    | Clausify th -> [th];;

let compareThm (Thm (fm1,_), Thm (fm2,_)) = compare (fm1,fm2);;

let parentsThm (Thm (_,inf)) = parentsInference inf;;

let mkAxiom fm = Thm (fromFormula fm, Axiom fm);;

let destThm (Thm (fm,inf)) = (toFormula fm, inf);;


  let emptyProved : (thm,Formula.formula) Pmap.map = Pmap.newMap compareThm;;

  let isProved proved th = Pmap.inDomain th proved;;

  let isUnproved proved th = not (isProved proved th);;

  let lookupProved proved th =
      match Pmap.peek proved th with
        Some fm -> fm
      | None -> raise (Bug "Normalize.lookupProved");;

  let rec prove acc proved ths =
      match ths with
        [] -> List.rev acc
      | th :: ths' ->
        if isProved proved th then prove acc proved ths'
        else
            let pars = parentsThm th

            in let deps = List.filter (isUnproved proved) pars
          in
            if Mlist.null deps then
                let (fm,inf) = destThm th

                in let fms = List.map (lookupProved proved) pars

                in let acc = (fm,inf,fms) :: acc

                in let proved = Pmap.insert proved (th,fm)
              in
                prove acc proved ths'
            else
                let ths = deps @ ths
              in
                prove acc proved ths
          ;;

  let proveThms = prove [] emptyProved;;

let toStringInference inf =
    match inf with
      Axiom _ -> "Axiom"
    | Definition _ -> "Definition"
    | Simplify _ -> "Simplify"
    | Conjunct _ -> "Conjunct"
    | Specialize _ -> "Specialize"
    | Skolemize _ -> "Skolemize"
    | Clausify _ -> "Clausify";;


(* ------------------------------------------------------------------------- *)
(* Simplifying with definitions.                                             *)
(* ------------------------------------------------------------------------- *)

type simplify =
      {formula : (formula, formula * thm) Pmap.map;
       andSet : (formula Pset.set * formula * thm) list;
       orSet : (formula Pset.set * formula * thm) list;
       xorSet : (formula Pset.set * formula * thm) list};;

let simplifyEmpty =
      {formula = Pmap.newMap compare;
       andSet = [];
       orSet = [];
       xorSet = []};;

  let simpler fm s =
      Pset.size s <> 1 or
      match Pset.pick s with
        True -> false
      | False -> false
      | Literal _ -> false
      | _ -> true;;

  let addSet set_defs body_def =
        let def_body_size (body,_,_) = Pset.size body

        in let body_size = def_body_size body_def

        in let (body,_,_) = body_def

        in let rec add acc = function
            [] -> Mlist.revAppend (acc,[body_def])
          | (((b,_,_) as bd) :: bds) as l ->
            match Int.compare (def_body_size bd, body_size) with
              Less -> Mlist.revAppend (acc, body_def :: l)
            | Equal ->
              if Pset.equal b body then Mlist.revAppend (acc,l)
              else add (bd :: acc) bds
            | Greater -> add (bd :: acc) bds
      in
        add [] set_defs
      ;;

  let rec add ({formula=formula;andSet=andSet;orSet=orSet;xorSet=xorSet} as simp) = function
      (body,False,th) -> add simp (negate body, True, th)
    | (True,_,_) -> simp
    | (And (_,_,s), def, th) ->
        let andSet = addSet andSet (s,def,th)
        and orSet = addSet orSet (negateSet s, negate def, th)
      in
          {formula = formula;
           andSet = andSet;
           orSet = orSet;
           xorSet = xorSet}
    | (Or (_,_,s), def, th) ->
        let orSet = addSet orSet (s,def,th)
        and andSet = addSet andSet (negateSet s, negate def, th)
      in
          {formula = formula;
           andSet = andSet;
           orSet = orSet;
           xorSet = xorSet}
    | (Xor (_,_,p,s), def, th) ->
        let simp = addXorSet simp (s, applyPolarity p def, th)
      in
        (match def with
          True ->
            let addXorLiteral = function
                (Literal _ as fm, simp) ->
                  let s = Pset.delete s fm
                in
                  if not (simpler fm s) then simp
                  else addXorSet simp (s, applyPolarity (not p) fm, th)
              | (_,simp) -> simp
          in
            Pset.foldl addXorLiteral simp s
        | _ -> simp)
    | (body,def,th) ->
      if Pmap.inDomain body formula then simp
      else
          let formula = Pmap.insert formula (body,(def,th))
          in let formula = Pmap.insert formula (negate body, (negate def, th))
        in
            {formula = formula;
             andSet = andSet;
             orSet = orSet;
             xorSet = xorSet}

  and addXorSet (({formula=formula;andSet=andSet;orSet=orSet;xorSet=xorSet}) as simp) (s,def,th) =
      if Pset.size s = 1 then add simp (Pset.pick s, def, th)
      else
          let xorSet = addSet xorSet (s,def,th)
        in
            {formula = formula;
             andSet = andSet;
             orSet = orSet;
             xorSet = xorSet}
        ;;

  let simplifyAdd simp (Thm (fm,_) as th) = add simp (fm,True,th);;


  let simplifySet set_defs set =
        let pred (s,_,_) = Pset.subset s set
      in
        match Mlist.find pred set_defs with
          None -> None
        | Some (s,f,th) ->
            let set = Pset.add (Pset.difference set s) f
          in
            Some (set,th)
      ;;

  let simplify ({formula=formula;andSet=andSet;orSet=orSet;xorSet=xorSet}) =
        let rec simp fm inf =
            match simp_sub fm inf with
              None -> simp_top fm inf
            | Some (fm,inf) -> try_simp_top fm inf

        and try_simp_top fm inf =
            match simp_top fm inf with
              None -> Some (fm,inf)
            | x -> x

        and simp_top fm inf =
            match fm with
              And (_,_,s) ->
              (match simplifySet andSet s with
                 None -> None
               | Some (s,th) ->
                   let fm = fAndSet s
                   in let inf = th :: inf
                 in
                   try_simp_top fm inf
                 )
            | Or (_,_,s) ->
              (match simplifySet orSet s with
                 None -> None
               | Some (s,th) ->
                   let fm = fOrSet s
                   in let inf = th :: inf
                 in
                   try_simp_top fm inf
                 )
            | Xor (_,_,p,s) ->
              (match simplifySet xorSet s with
                 None -> None
               | Some (s,th) ->
                   let fm = fXorPolaritySet (p,s)
                   in let inf = th :: inf
                 in
                   try_simp_top fm inf
                 )
            | _ ->
              (match Pmap.peek formula fm with
                 None -> None
               | Some (fm,th) ->
                   let inf = th :: inf
                 in
                   try_simp_top fm inf
                 )

        and simp_sub fm inf =
            match fm with
              And (_,_,s) ->
              (match simp_set s inf with
                 None -> None
               | Some (l,inf) -> Some (fAndList l, inf))
            | Or (_,_,s) ->
              (match simp_set s inf with
                 None -> None
               | Some (l,inf) -> Some (fOrList l, inf))
            | Xor (_,_,p,s) ->
              (match simp_set s inf with
                 None -> None
               | Some (l,inf) -> Some (fXorPolarityList (p,l), inf))
            | Exists (_,_,n,f) ->
              (match simp f inf with
                 None -> None
               | Some (f,inf) -> Some (fExistsSet (n,f), inf))
            | Forall (_,_,n,f) ->
              (match simp f inf with
                 None -> None
               | Some (f,inf) -> Some (fForallSet (n,f), inf))
            | _ -> None

        and simp_set s inf =
              let (changed,l,inf) = Pset.foldr simp_set_elt (false,[],inf) s
            in
              if changed then Some (l,inf) else None

        and simp_set_elt (fm,(changed,l,inf)) =
            match simp fm inf with
              None -> (changed, fm :: l, inf)
            | Some (fm,inf) -> (true, fm :: l, inf)
      in
        fun (Thm (fm,_) as th) ->
           match simp fm [] with
             Some (fm,ths) ->
               let inf = Simplify (th,ths)
             in
               Thm (fm,inf)
           | None -> th
      ;;

(*MetisTrace2
let simplify = fun simp -> fun th as Thm (fm,_) ->
    let
      let th' as Thm (fm',_) = simplify simp th
      let () = if compare (fm,fm') = Equal then ()
               else (Print.trace pp "Normalize.simplify: fm" fm;;
                     Print.trace pp "Normalize.simplify: fm'" fm')
    in
      th'
    end;;
*)

(* ------------------------------------------------------------------------- *)
(* Definitions.                                                              *)
(* ------------------------------------------------------------------------- *)

  let dRCounter : int ref = ref 0;;

  let newDR () =
        let i = !dRCounter
        in let () = dRCounter := i + 1
      in
        definitionPrefix ^ "_" ^ Int.toString i
      ;;

  let newDefinitionRelation () = Portable.critical newDR ();;


let newDefinition def =
      let fv = freeVars def
      in let rel = newDefinitionRelation ()
      in let atm = (Name.fromString rel, Name.Set.transform (fun x -> Term.Var x) fv)
      in let fm = Formula.Iff (Formula.Atom atm, toFormula def)
      in let fm = Formula.setMkForall (fv,fm)
      in let inf = Definition (rel,fm)
      in let lit = Literal (fv,(false,atm))
      in let fm = fXor2 (lit,def)
    in
      Thm (fm,inf)
    ;;

(* ------------------------------------------------------------------------- *)
(* Definitional conjunctive normal form.                                     *)
(* ------------------------------------------------------------------------- *)

type cnf =
    Consistent_cnf of simplify
  | Inconsistent_cnf;;

let initialCnf = Consistent_cnf simplifyEmpty;;


  let def_cnf_inconsistent th =
        let cls = [(Literal.Set.empty,th)]
      in
        (cls,Inconsistent_cnf)
      ;;

  let def_cnf_clause inf (fm,acc) =
        let cl = toClause fm
        in let th = Thm (fm,inf)
      in
        (cl,th) :: acc
      ;;
(*MetisDebug
      handle Error err ->
        (Print.trace pp "Normalize.addCnf.def_cnf_clause: fm" fm;;
         raise (Bug ("Normalize.addCnf.def_cnf_clause: " ^ err)));;
*)

  let rec def_cnf cls simp ths =
      match ths with
        [] -> (cls, Consistent_cnf simp)
      | th :: ths -> def_cnf_formula cls simp (simplify simp th) ths

  and def_cnf_formula cls simp ((Thm (fm,_)) as th) ths =
      match fm with
        True -> def_cnf cls simp ths
      | False -> def_cnf_inconsistent th
      | And (_,_,s) ->
          let add (f,z) = Thm (f, Conjunct th) :: z
        in
          def_cnf cls simp (Pset.foldr add ths s)
      | Exists (fv,_,n,f) ->
          let th = Thm (skolemize fv n f, Skolemize th)
        in
          def_cnf_formula cls simp th ths
      | Forall (_,_,_,f) ->
          let th = Thm (f, Specialize th)
        in
          def_cnf_formula cls simp th ths
      | _ ->
        match minimumDefinition fm with
          Some def ->
            let ths = th :: ths
            in let th = newDefinition def
          in
            def_cnf_formula cls simp th ths
        | None ->
            let simp = simplifyAdd simp th

            in let fm = basicCnf fm

            in let inf = Clausify th
          in
            match fm with
              True -> def_cnf cls simp ths
            | False -> def_cnf_inconsistent (Thm (fm,inf))
            | And (_,_,s) ->
                let inf = Conjunct (Thm (fm,inf))
                in let cls = Pset.foldl (def_cnf_clause inf) cls s
              in
                def_cnf cls simp ths
            | fm -> def_cnf (def_cnf_clause inf (fm,cls)) simp ths
          ;;

  let addCnf th cnf =
      match cnf with
        Consistent_cnf simp -> def_cnf [] simp [th]
      | Inconsistent_cnf -> ([],cnf);;


  let add (th,(cls,cnf)) =
        let (cls',cnf) = addCnf th cnf
      in
        (cls' @ cls, cnf)
      ;;
  let proveCnf ths =
        let (cls,_) = Mlist.foldl add ([],initialCnf) ths
      in
        List.rev cls
      ;;



let cnf fm =
      let cls = proveCnf [mkAxiom fm]
    in
      List.map fst cls
    ;;
