(* The definitions of the basic types for the parsing, building and
   printing of TPTP and extended TPTP formats *)

(* First order terms *)
type fof_tm =
   V of int
 | A of string * fof_tm list;;

(* We choose to express an implication 'a=>b' as '~a | b',
   however biimplications may be transformed in various ways, so are kept *)
type fof_form =
   Atom of (string * fof_tm list)
 | Neg of fof_form
 | Conj of fof_form * fof_form
 | Disj of fof_form * fof_form
 | Forall of int * fof_form
 | Exists of int * fof_form
 | Eqiv of fof_form * fof_form;;

(* In CNF variables do not need to be quantified.
   We just need to assign them distinct numbers *)
let var =
  let var_no, no_var, var_num = Hashtbl.create 100, Hashtbl.create 100, ref 0 in
  fun (name : string) ->
    try Hashtbl.find var_no name with Not_found ->
    incr var_num; Hashtbl.add var_no name !var_num; Hashtbl.add no_var !var_num name; !var_num;;

let fof_list_forall vs t = List.fold_right (fun v sf -> Forall (v, sf)) vs t;;
let fof_list_exists vs t = List.fold_right (fun v sf -> Exists (v, sf)) vs t;;

(* The type of general type-theory terms *)
type tt_tm =
   Id of string                          (* may be a constant or variable *)
 | Comb of tt_tm * tt_tm
 | Abs of string * tt_tm * tt_tm;;     (* name of introduced id, type and subterm *)

open Format;;
let arity = Hashtbl.create 1000;;
Hashtbl.add arity "=" 2;;

let save_arity (f, l) =
  let ll = List.length l in
  try if Hashtbl.find arity f <> ll then eprintf "Different arities for %s, things likely to break!\n" f
  with Not_found -> Hashtbl.add arity f ll;;

let pp_print_var f i =
  pp_print_char f (Char.chr (65 + i mod 26)); if i > 25 then pp_print_int f (i / 26);;

let rec pp_iter f fn sep = function
    [] -> ()
  | [e] -> fn f e
  | h :: t -> fn f h; pp_print_string f sep; pp_iter f fn sep t;;

let rec pp_print_term f = function
    V i -> pp_print_var f i
  | A (i, l) ->
      pp_print_string f (if Char.code i.[0] >= 48 && Char.code i.[0] <= 57 then ("number" ^ i) else i);
      if l <> [] then begin pp_print_char f '('; pp_iter f pp_print_term "," l; pp_print_char f ')' end;;

let rec pp_print_form f = function
  | Atom ("=",[t1;t2]) -> pp_print_term f t1; pp_print_string f "="; pp_print_term f t2
  | Atom (s,t) -> pp_print_term f (A (s,t))
  | Neg (Atom ("=",[t1;t2])) -> pp_print_term f t1; pp_print_string f "!="; pp_print_term f t2
  | Neg (Atom (s,t)) -> pp_print_string f "~"; pp_print_term f (A (s, t))
  | Neg t -> pp_print_string f "~("; pp_print_form f t; pp_print_char f ')'
  | Conj (l, r) -> pp_print_char f '('; pp_print_form f l; pp_print_string f "&"; pp_print_form f r; pp_print_char f ')'
  | Disj (Neg l, r) -> pp_print_char f '('; pp_print_form f l; pp_print_string f "=>"; pp_print_form f r; pp_print_char f ')'
  | Disj (l, r) -> pp_print_char f '('; pp_print_form f l; pp_print_string f "|"; pp_print_form f r; pp_print_char f ')'
  | Forall (v, t) -> pp_print_string f "!["; pp_print_var f v; pp_print_string f "]:"; pp_print_form f t
  | Exists (v, t) -> pp_print_string f "?["; pp_print_var f v; pp_print_string f "]:"; pp_print_form f t
  | Eqiv (l, r) -> pp_print_char f '('; pp_print_form f l; pp_print_string f "<=>"; pp_print_form f r; pp_print_char f ')'
;;
